import librds

def test_ps():
    assert librds.GroupInterface.getPS("radio95") == ("radio95 ", 4)

def test_ps_long():
    assert librds.GroupInterface.getPS("radio95 radio95") == ("radio95 ", 4)

def test_rt():
    assert librds.GroupInterface.getRT("hi!") == ("hi!\r", 1)

def test_rt2():
    assert librds.GroupInterface.getRT("hello!") == ("hello!\r ", 2)

def test_rt3():
    assert librds.GroupInterface.getRT("".ljust(64,".")) == ("".ljust(64,"."), 16)

def test_rt_full():
    assert librds.GroupInterface.getRT("hello!",True) == ("hello!".ljust(64), 16)

def test_ptyn():
    assert librds.GroupInterface.getPTYN("Various") == ("Various".ljust(8), 2)

def test_ptyn_long():
    assert librds.GroupInterface.getPTYN("Various and blah blah blah blah") == ("Various ".ljust(8), 2)


from enum import Enum, IntEnum

def get_from_list(input:list,index:int,default=None):
    """This is a simple function to remove index errors from the group generators"""
    try:
        return input[index]
    except IndexError:
        return default

def SubstituteCharacterAtPosition(string: str, char: str, index: int) -> str:
    """
    This is a helper function that works like 'string[index] = char' in C.

    :param string: String to process
    :param char: The character that is going to be at that position
    :param index: Index of the character
    :return: Modified string with the character at the specified index
    """
    if not isinstance(string, str) or not isinstance(char, str) or not isinstance(index, int):
        raise Exception("Invalid input types")
    if not (0 <= index < len(string)):
        raise IndexError("Index out of range")
    return string[:index] + char + string[index + 1:]

class BitManipulator:
    """
    This is a alternative to bitwising as its too complex to understand for me, so what did i do?
    This class can set/get bits of a number it a specific location using bin() and int(str,2)
    """
    @staticmethod
    def set_bit(value:int, bit:int, bit_value:bool, max_bits:int=16) -> int:
        """
        This function sets a single bit in a value

        :param value: Input int
        :param bit: The bit index to set
        :param bit_value: 0 or 1?
        :param max_bits: If you don't want unexpected indexerrors, then you can use this, this will pad the start of the bin string as 01 and 0000001 are the same thing
        """
        string = ("0"*(max_bits-(value.bit_length())))
        string += bin(value).removeprefix("-").removeprefix("0b")
        
        if bit > len(string): raise IndexError
        string = SubstituteCharacterAtPosition(string,("1" if bit_value else "0"),bit)
        return int(("0b"+string),2)
    @staticmethod
    def get_bit(value:int, bit:int, max_bits:int=16) -> int:
        string = ("0"*(max_bits-(value.bit_length())))
        string += bin(value).removeprefix("-").removeprefix("0b")
        return int(("0b"+string[bit]),2)
    @staticmethod
    def set_bits(value: int, bits: list[int], bit_values: list[int], max_bits: int = 16) -> int:
        print(bit_values)
        out = value
        for i, bit_pos in enumerate(bits):
            out = BitManipulator.set_bit(out, bit_pos, bit_values[i], max_bits)
        return out
    @staticmethod
    def get_bits(value: int, bits: int, index:int=0, max_bits:int=16) -> int:
        out = "0b"
        for i in range(bits):
            out += bin(BitManipulator.get_bit(value, (i+index), max_bits)).removeprefix("0b")
        return int(out,2)

def calculate_mjd(year: int, month: int, day:int):
    """
    :param year: e.x 2024
    :param month: (starts from 0) 0 - Jan, 1 - Feb ...
    :param day: (starts from 1)"""
    l = 1 if (month == 0 or month == 1) else 0
    return (
        14956 + day + 
        int(
            ((year - 1900) - l) * 365.25
        ) +
        int(
            (month + 2 + l * 12) * 30.6001
        )
    )

def calculate_ymd(mjd:int):
    """Returns the same format as calculate_mjd, so you can encode and decode without any conversions"""
    jd = mjd + 2_400_001
    ljd = jd + 68569
    
    njd = int((4 * ljd / 146097))
    ljd = ljd - int(((146097 * njd + 3) / 4))
    
    year = int((4000 * (ljd + 1) / 1461001))
    ljd = ljd - int(((1461 * year / 4))) + 31
    
    month = int((80 * ljd / 2447))
    
    day = ljd - int((2447 * month / 80))
    
    ljd = int((month / 11))
    month = int((month + 2 - 12 * ljd))
    year = int((100 * (njd - 49) + year + ljd))
    return year, (month-1), day

class Groups(Enum):
    PS = 0
    PS_B = 1
    RT = 2
    RT_B = 3
    PTYN = 4
    ECC = 5
    LIC = 6
    TDC = 7
    TDC_B = 8
    IN_HOUSE = 9
    IN_HOUSE_B = 10

class GroupSequencer:
    """you can use this code to sequence though anything"""
    def __init__(self, sequence:list[IntEnum]) -> None:
        self.cur_idx = 1
        self.sequence = sequence
    def get_next(self):
        if len(self.sequence) == 0: return
        if self.cur_idx > len(self.sequence): self.cur_idx = 1
        prev = self.sequence[self.cur_idx-1]
        self.cur_idx += 1
        return prev
    def change_sequence(self, sequence:list[IntEnum]):
       self.sequence = sequence
       self.cur_idx = 1
    def __len__(self):
        return len(self.sequence)
    def __iter__(self):
        return self.sequence



import librds

def test_sequencer():
    input = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
    seq = librds.GroupSequencer(input)
    a = []
    for _ in range(len(input)*2):
        a.append(seq.get_next())
    # We circled it 2 times so 2 have input+input
    assert a == (input+input)


from .interface import GroupInterface
from .comfort import Groups, GroupSequencer, calculate_mjd, calculate_ymd
from .af import AF_Bands, AlternativeFrequencyEntry, AlternativeFrequency
from .generator import GroupGenerator, Group, GroupIdentifier
from .decoder import GroupDecoder
librds_version = 1.7
__version__ = librds_version



from .datatypes import Group, GroupIdentifier
from .comfort import get_from_list, BitManipulator
from .af import AlternativeFrequency
from .charset import RDSCharset
class SegmentError(Exception): pass
class GroupGeneratorError(Exception): pass

def get_group(group:int, ab:bool=False,details:bool=False):
    group = ((group << 4) | int(ab))
    if not details: 
        return (((group >> 4) & 15) << 12) | (0x800 if ab else 0)
    else:
        return ((group >> 4) & 15), int(ab)


class GroupGenerator:
    def basic(pi:int, tp:bool=False, pty:int=0):
        """
        This function will generate a basic block structure which includes the PI, TP and PTY, this shouldn't be sent by itself to a decoder
        
        :param pi: The PI code, should be a 16 bit integer (for example 0x3000), pi codes are country dependent (especially for ECC), see Standart EN50067 page 70 (first byte, so 3 in 0x3555), also see page 72 for coverage code
        :param tp: Traffic Program
        :param pty: Program Type, see EN50067 page 77-80
        """
        return Group(
            pi & 0xFFFF, #A
            (int(tp) << 10 | pty << 5) & 0xFFFF, #B
            0,0, #C
            None #is_b
        )
    def ps(blocks:Group, ps_text:str, segment:int, ms:bool=True, ta:bool=False, di:int=1, block2:int=0):
        """
        This function will generate a 0A group which includes TA MS DI AF and 2 characters of text
        
        :param blocks: Basic block structure, which is generated by basic()
        :param ps_text: PS Text
        :param segment: Current segment to generate (ps is segmented into four 2 character segments)
        :param ms: Music/Speech (music is True)
        :param ta: Trafic Annoucement
        :param di: Decoder Identificaton
        :param block2: Block C, you can use this to send AF

        Decoder status: Decoded
        """
        if segment > 3: raise SegmentError("Segment limit")
        return Group(
            blocks.a & 0xFFFF,
            ( blocks.b | int(ta) << 4 | int(ms) << 3 | ( ( di >> (3-segment) ) << 2) | segment ) & 0xFFFF,
            (int(block2) & 0xFFFF) or AlternativeFrequency.get_no_af(),
            (RDSCharset.translate(get_from_list(ps_text,segment*2," "))<<8 | RDSCharset.translate(get_from_list(ps_text,segment*2+1," "))) & 0xFFFF, #low byte + high byte as we're using utf-8, and we conver 2 8 bit numbers to a single 16 bit one
            False
        )
    def ps_b(blocks:Group,ps_text:str,segment:int,ms:bool=True,ta:bool=False,di:int=1):
        """This function will generate a 0B group which is very similiar to a 0A group but no AF
        
        :param blocks: Basic block structure, which is generated by basic()
        :param ps_text: PS Text
        :param segment: Current segment to generate (ps is segmented into four 2 character segments)
        :param ms: Music/Speech (music is True)
        :param ta: Trafic Annoucement
        :param di: Decoder Identificaton

        Decoder status: Decoded
        """
        if segment > 3: raise SegmentError("Segment limit")
        return Group(
            blocks.a & 0xFFFF,
            ( blocks.b | get_group(0,True) | int(ta) << 4 | int(ms) << 3 | ( ( di >> (3-segment) ) << 2) | segment ) & 0xFFFF,
            blocks.a & 0xFFFF,
            (RDSCharset.translate(get_from_list(ps_text,segment*2," "))<<8 | RDSCharset.translate(get_from_list(ps_text,segment*2+1," "))) & 0xFFFF,
            True
        )
    def ecc(blocks:Group, ecc: int):
        """This function will generate a 1A group, with the ECC value
        
        :param blocks: Basic block structure, which is generated by basic()
        :param ecc: The ECC code itself, see EN50067 page 71

        Decoder status: Decoded
        """
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | get_group(1)) & 0xFFFF,
            (blocks.c | ecc) & 0xFFFF,
            blocks.d & 0xFFFF,
            False
        )
    def lic(blocks:Group, lic: int):
        """This function will generate a 1A group, very similiar to ECC
        
        :param blocks: Basic block structure, which is generated by basic()
        :param lic: The LIC code itself, see EN50067 page 84-85

        Decoder status: Decoded
        """
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | get_group(1)) & 0xFFFF,
            (blocks.c | (lic | 0x3000)) & 0xFFFF,
            blocks.d & 0xFFFF,
            False
        )
    def rt(blocks:Group,rt_text:str,segment:int,ab:bool=False):
        """This function will generate a 2A group, which includes AB the segment and 4 characters of text
        
        :param blocks: Basic block structure, which is generated by basic()
        :param rt_text: RT Text
        :param segment: Current segment to generate (2a is segmented into 16 4 character segments)
        :param ab: Switch this if RT changes

        Decoder status: Decoded
        """
        if segment > 15: raise SegmentError("Segment limit")
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | get_group(2) | int(ab) << 4 | segment) & 0xFFFF,
            (RDSCharset.translate(get_from_list(rt_text,segment*4+0," "))<<8 | RDSCharset.translate(get_from_list(rt_text,segment*4+1," "))) & 0xFFFF,
            (RDSCharset.translate(get_from_list(rt_text,segment*4+2," "))<<8 | RDSCharset.translate(get_from_list(rt_text,segment*4+3," "))) & 0xFFFF,
            False
        )
    def rt_b(blocks:Group,rt_text:str,segment:int,ab:bool=False):
        """This function will generate a 2B group, very similiar to 2A but this include only 2 characters of text
        
        :param blocks: Basic block structure, which is generated by basic()
        :param rt_text: RT Text
        :param segment: Current segment to generate (2b is segmented into 16 2 character segments)
        :param ab: Switch this if RT changes

        Decoder status: Decoded
        """
        if segment > 15: raise SegmentError("Segment limit")
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | get_group(2,True) | int(ab) << 4 | segment) & 0xFFFF,
            blocks.a & 0xFFFF,
            (RDSCharset.translate(get_from_list(rt_text,segment*4+0," "))<<8 | RDSCharset.translate(get_from_list(rt_text,segment*4+1," "))) & 0xFFFF,
            True
        )
    def ptyn(blocks:Group, ptyn_text:str, segment:int,ab:bool=False):
        """This function will generate a 10A group, which includes 4 characters of text
        
        :param blocks: Basic block structure, which is generated by basic()
        :param ptyn_text: PTYN Text
        :param segment: Current segment to generate (2b is segmented into 2 2 character segments)
        :param ab: Switch this if PTYN changes

        Decoder status: Decoded
        """
        if segment > 1: raise SegmentError("Segment limit")
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | get_group(10) | (int(ab) << 4) | segment) & 0xFFFF,
            (RDSCharset.translate(get_from_list(ptyn_text,segment*4+0," "))<<8 | RDSCharset.translate(get_from_list(ptyn_text,segment*4+1," "))) & 0xFFFF,
            (RDSCharset.translate(get_from_list(ptyn_text,segment*4+2," "))<<8 | RDSCharset.translate(get_from_list(ptyn_text,segment*4+3," "))) & 0xFFFF,
            False
        )
    def ct(basic:Group, mjd: int, hour: int, minute: int, local_hour:int=None,local_minute:int=None):
        """This function generates a 4A group
        
        :param blocks: Basic block structure, which is generated by basic()
        :param mjd: The Modified Julian Day value, calculated with calculate_mjd
        :param hour: UTC Hour
        :param minute: UTC Minute
        :param local_hour: Local Hour
        :param local_minute: Local Minute

        Decoder status: Decoded
        """
        group = Group(
            basic.a & 0xFFFF,
            (basic.b | get_group(4) | (mjd>>15)) & 0xffff,
            ((mjd<<1) | (hour >> 4)) & 0xffff,
            ((hour & 0xF)<<12 | minute << 6) & 0xffff,
            False
        )
        if local_hour and local_minute:
            offset_h = local_hour - hour
            if offset_h < 0: group.d |= (1 << 5)
            if local_minute:
                offset_m = local_minute - minute
            else: offset_m = 0
            offset = int((offset_m + (offset_h * 60)) / 30)
            group.d |= abs(offset)
        elif local_hour:
            offset_h = local_hour - hour
            if offset_h < 0: group.d |= (1 << 5)
            offset = int((0 + (offset_h * 60)) / 30)
            group.d |= abs(offset)
        elif local_minute:
            offset_m = local_minute - minute
            if offset_m < 0: group.d |= (1 << 5)
            offset = int((offset_m + (0 * 60)) / 30)
            group.d |= abs(offset)
        return group
    def tdc(blocks:Group, channel:int,data:list[int],segment:int):
        """This function will generate a 5A group
        
        :param blocks: Basic block structure, which is generated by basic()
        :param channnel: TDC channel< this should stay static, shouldn't be 31 ot more (32 but it starts from 0 so 31)
        :param data: TDC Data, should be a list of ints
        :param segment: Segment of the TDC message

        Decoder status: Decoded
        """
        if channel > 31: raise GroupGeneratorError("Channel bigger than 31")
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | get_group(5) | (channel & 0x1F)) & 0xFFFF,
            (get_from_list(data,segment+0,0)<<8 | get_from_list(data,segment+1,0)) & 0xFFFF,
            (get_from_list(data,segment+2,0)<<8 | get_from_list(data,segment+3,0)) & 0xFFFF,
            False
        )
    def tdc_b(blocks:Group, channel:int,data:list[int],segment:int):
        """This function will generate a 5B group
        
        :param blocks: Basic block structure, which is generated by basic()
        :param channnel: TDC channel< this should stay static, shouldn't be 31 ot more (32 but it starts from 0 so 31)
        :param data: TDC Data, should be a list of ints
        :param segment: Segment of the TDC message

        Decoder status: Decoded
        """
        if channel > 31: raise GroupGeneratorError("Channel bigger than 31")
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | get_group(5,True) | (channel & 0x1F)) & 0xFFFF,
            blocks.a & 0xFFFF,
            (get_from_list(data,segment+0,0)<<8 | get_from_list(data,segment+1,0)) & 0xFFFF,
            True
        )
    def in_house(blocks:Group, data:list[int],segment:int=0):
        """This function will generate a 6A group
        
        :param blocks: Basic block structure, which is generated by basic()
        :data data: 3 byte long list of ints, but the 1st byte is not guarteed to be decoded so maybe but some unnessacy info here

        Decoder status: Decoded
        """
        return Group(
            blocks.a & 0xFFFF,
            (((blocks.b | get_group(6)) & ~0x1F) | (get_from_list(data,segment+0,0) & 0x1F)) & 0xFFFF,
            get_from_list(data,segment+1,0) & 0xFFFF,
            get_from_list(data,segment+2,0) & 0xFFFF,
            False
        )
    def in_house_b(blocks:Group, data:list[int],segment:int=0):
        """This function will generate a 6B group
        
        :param blocks: Basic block structure, which is generated by basic()
        :data data: 2 byte long list of ints

        Decoder status: Decoded
        """
        return Group(
            blocks.a & 0xFFFF,
            (((blocks.b | get_group(6,True)) & ~0x1F) | (get_from_list(data,segment+0,0) & 0x1F)) & 0xFFFF,
            blocks.a & 0xFFFF,
            get_from_list(data,segment+1,0) & 0xFFFF,
            True
        )
    def oda_aid(blocks:Group, group: GroupIdentifier, aid:int, scb:int=0):
        """This function will generate a 3A group
        
        :param blocks: Basic block structure, which is generated by basic()
        :data group: The type of group this AID is for
        :data aid: AID
        :data scb: 'Message bits'

        Decoder status: Decoded
        """
        if not isinstance(group, GroupIdentifier): raise GroupGeneratorError("group should be a GroupIdentifier")
        group_type, group_ver = get_group(group.group_number,group.group_version,True)
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | 3 << 12 | (group_type << 1) | group_ver) & 0xFFFF,
            scb & 0xFFFF,
            aid & 0xFFFF,
            False
        )
    def oda(blocks:Group, group: GroupIdentifier, data:list[int,int,int]):
        """This function will generate a xY group
        
        :param blocks: Basic block structure, which is generated by basic()
        :data group: The type of group this ODA is
        :data data: Raw 16 bit data blocks (first is 5 bits)

        Decoder status: Won't decode
        """
        if not isinstance(group, GroupIdentifier): raise GroupGeneratorError("group should be a GroupIdentifier")
        return Group(
            blocks.a & 0xFFFF,
            (((blocks.b | get_group(group.group_number,group.group_version)) & ~0x1F) | get_from_list(data,0,0) & 0x1F) & 0xFFFF,
            get_from_list(data,1,0) & 0xFFFF,
            get_from_list(data,2,0) & 0xFFFF,
            False
        )
    def eon(blocks:Group, on_pi: int, on_tp: bool, variant_code: int, on_ps:str="", on_af:int=0, on_pty: int=0, on_ta: bool=False):
        """This function will generate a 14A group (UNTESTED)
        
        :param blocks: Basic block structure, which is generated by basic()
        :param on_pi: The other station's pi
        :param on_tp: The other station's tp
        :param variant_code: The variant code, see http://www.interactive-radio-system.com/docs/EN50067_RDS_Standard.pdf page 38
        :param on_ps: Other station's data
        :param on_af: Other station's data
        :param on_pty: Other station's data
        :param on_ta: Other station's data
        Decoder status: Decoded
        """
        group = Group(
            blocks.a & 0xFFFF,
            (((blocks.b | get_group(14,False) | BitManipulator.set_bit(0,11,on_tp)) & ~0xF) | (variant_code & 0xF)) & 0xFFFF,
            0,
            on_pi,
            False
        )
        if variant_code <= 3:
            group.c = (RDSCharset.translate(get_from_list(on_ps,variant_code*2+0," ")) << 8) | RDSCharset.translate(get_from_list(on_ps,variant_code*2+1," "))
        match variant_code:
            case 4:
                group.c = on_af
            case 13:
                group.c = ((on_pty & 0x1f) << 11) | (int(on_tp) << 10)
        return group
    def eon_b(blocks:Group, on_pi: int, on_tp: bool, on_ta:bool):
        """This function will generate a 14B group (UNTESTED)
        
        :param blocks: Basic block structure, which is generated by basic()
        :param on_pi: The other station's pi
        :param on_tp: The other station's tp
        :param on_tp: The other station's ta
        Decoder status: Decoded
        """
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | get_group(14,True) | BitManipulator.set_bit(0,11,on_tp) | BitManipulator.set_bit(0,12,on_ta)) & 0xFFFF,
            blocks.a & 0xFFFF,
            on_pi & 0xFFFF,
            True
        )


from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as f:
    readme = f.read()

setup(
        name="librds", 
        version="1.7",
        author="kuba201",
        description='RDS Group Generator',
        long_description=readme,
        long_description_content_type='text/markdown',
        packages=find_packages(),
        url="https://flerken.zapto.org:1115/kuba/librds",
        install_requires=[],
        project_urls={
            'Source': 'https://flerken.zapto.org:1115/kuba/librds',
        },
        keywords=['radiodatasystem','rds','broadcast_fm'],
        classifiers= [
            "Intended Audience :: Education",
            "Intended Audience :: Telecommunications Industry",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.10",
            "Development Status :: 5 - Production/Stable",
            "License :: OSI Approved :: GNU General Public License (GPL)",
        ]
)


from .datatypes import Group, DecodedGroup, GroupIdentifier, PSDetails, RTDetails, ECCLICDetails, PTYNDetails, CTDetails, TDCDetails, InHouseDetails, ODAAidDetails, EONBDetails, EONADetails
from .charset import RDSCharsetDecode
from .comfort import BitManipulator

class GroupDecoder:
    def decode(group:Group):
        def readValue(val, bits, index):
            return BitManipulator.get_bits(val,bits,index)
        
        def readGroup():
            return readValue(group.b,4,0), BitManipulator.get_bit(group.b, 4)
        def readTP():
            return bool(BitManipulator.get_bit(group.b,5))
        def readPTY():
            return readValue(group.b, 5, 6)
        group_number, group_version = readGroup()
        group_out = DecodedGroup(group,group.a,readTP(),readPTY(), GroupIdentifier(group_number, group_version), None)
        
        def decode_group_0():
            def readTAMS():
                return bool(BitManipulator.get_bit(group.b,11)), bool(BitManipulator.get_bit(group.b,12))
            def readSegment():
                is_di = bool(BitManipulator.get_bit(group.b,13))
                return is_di, readValue(group.b, 2, 14)
            ta, ms = readTAMS()
            is_di, segment = readSegment()
            det = PSDetails(segment,None,None,None,None,ms,ta,"")
            #i've spent a long time understanding di, and to be honest when i understand it, its fucking genius (from redsea's source code)
            match segment:
                case 0:
                    det.di_dpty = is_di
                case 1:
                    det.di_compressed = is_di
                case 2:
                    det.di_artificial_head = is_di
                case 3:
                    det.di_stereo = is_di
            char_0 = (group.d >> 8) & 0xFF
            char_1 = group.d & 0xFF
            det.text += RDSCharsetDecode.translate(char_0)
            det.text += RDSCharsetDecode.translate(char_1)
            group_out.details = det
        def decode_group_2():
            def readABSegment():
                return bool(BitManipulator.get_bit(group.b,11)), readValue(group.b, 4, 12)
            ab, segment = readABSegment()
            det = RTDetails(segment,ab,"")
            if not group_out.group.group_version == 1:
                char_1 = (group.c >> 8) & 0xFF
                char_2 = group.c & 0xFF
                char_3 = (group.d >> 8) & 0xFF
                char_4 = group.d & 0xFF
                det.text += RDSCharsetDecode.translate(char_1)
                det.text += RDSCharsetDecode.translate(char_2)
                det.text += RDSCharsetDecode.translate(char_3)
                det.text += RDSCharsetDecode.translate(char_4)
            else:
                char_1 = (group.d >> 8) & 0xFF
                char_2 = group.d & 0xFF
                det.text += RDSCharsetDecode.translate(char_1)
                det.text += RDSCharsetDecode.translate(char_2)
            group_out.details = det
        def decode_group_1():
            det = ECCLICDetails(group.c,None)
            det.is_lic = (bool(BitManipulator.get_bit(group.c,2)) and bool(BitManipulator.get_bit(group.c,3)))
            group_out.details = det
        def decode_group_10():
            def readABSegment():
                return bool(BitManipulator.get_bit(group.b,11)), int(BitManipulator.get_bit(group.b,15))
            ab, segment = readABSegment()
            det = PTYNDetails(segment,ab,"")
            char_1 = (group.c >> 8) & 0xFF
            char_2 = group.c & 0xFF
            char_3 = (group.d >> 8) & 0xFF
            char_4 = group.d & 0xFF
            det.text += RDSCharsetDecode.translate(char_1)
            det.text += RDSCharsetDecode.translate(char_2)
            det.text += RDSCharsetDecode.translate(char_3)
            det.text += RDSCharsetDecode.translate(char_4)
            group_out.details = det
        def decode_group_4():
            mjd_buffer = "0b"
            mjd_buffer += str(BitManipulator.get_bit(group.b,14))
            mjd_buffer += str(BitManipulator.get_bit(group.b,15))
            mjd_buffer += bin(readValue(group.c,15,0)).removeprefix("0b")
            hour_buffer = "0b"
            hour_buffer += str(BitManipulator.get_bit(group.c,15))
            hour_buffer += bin(readValue(group.d,4,0)).removeprefix("0b")
            
            mjd = int(mjd_buffer,2)
            hour = int(hour_buffer,2)
            minute = readValue(group.d,6,4)
            time_sense = bool(BitManipulator.get_bit(group.d,10))
            time_diffence = readValue(group.d,5,11)
            group_out.details = CTDetails(mjd, hour, minute, time_sense, time_diffence)
        def decode_group_5():
            address = readValue(group.b,5,11)
            data = []
            if group_version == 0:
                data.append((group.c >> 8) & 0xFF)
                data.append((group.c & 0xFF))
            data.append((group.d >> 8) & 0xFF)
            data.append((group.d & 0xFF))
            group_out.details = TDCDetails(address, data)
        def decode_group_6():
            data_from_b = readValue(group.b,5,11)
            data = [data_from_b]
            if group_version == 0:
                data.append(group.c)
            data.append(group.d)
            group_out.details = InHouseDetails(data)
        def decode_group_3():
            group_number_odaaid = readValue(group.b, 4, 11)
            group_version_odaaid = bool(BitManipulator.get_bit(group.b,15))
            group_out.details = ODAAidDetails(GroupIdentifier(group_number_odaaid, group_version_odaaid), group.d, group.c)
        def decode_group_14():
            on_pi = group.d
            on_tp = bool(BitManipulator.get_bit(group.b,10))
            if group.is_version_b:
                on_ta = bool(BitManipulator.get_bit(group.b,11))
                group_out.details = EONBDetails(on_pi, on_tp, on_ta)
            else:
                variant_code = readValue(group.b,4,12)
                on_ps_segment, on_ps_text = None, ""
                on_af, on_pty, on_ta = None, None, None
                if variant_code < 4:
                    on_ps_segment = variant_code
                    char_1 = (group.c >> 8) & 0xFF
                    char_2 = group.c & 0xFF
                    on_ps_text += RDSCharsetDecode.translate(char_1)
                    on_ps_text += RDSCharsetDecode.translate(char_2)
                match variant_code:
                    case 4:
                        on_af = group.c
                    case 13:
                        on_pty = readValue(group.c,5,0)
                        on_ta = bool(BitManipulator.get_bit(group.c,15))
                group_out.details = EONADetails(on_pi, on_tp, on_ta, on_ps_segment, on_ps_text, on_pty, on_af, variant_code)
            

        match group_number:
            case 0:
                decode_group_0()
            case 1:
                if group_version == 0: decode_group_1()
            case 2:
                decode_group_2()
            case 3:
                if group_version == 0: decode_group_3()
            case 4:
                if group_version == 0: decode_group_4()
            case 5:
                decode_group_5()
            case 6:
                decode_group_6()
            case 10:
                if group_version == 0: decode_group_10()
            case 14:
                decode_group_14()
        return group_out